# stdlib
from typing import List

# langchain
from langchain.agents.agent_toolkits.base import BaseToolkit
from langchain.tools import BaseTool

# langchain typing
from langchain.schema.language_model import BaseLanguageModel
from langchain.schema.vectorstore import VectorStore

# sigmaiq tools
from sigmaiq.llm.tools.create_sigma_rule import CreateSigmaRuleVectorStoreTool
from sigmaiq.llm.tools.translate_sigma_rule import TranslateSigmaRuleTool
from sigmaiq.llm.tools.find_sigma_rule import FindSigmaRuleTool
from sigmaiq.llm.tools.query_to_sigma_rule import QueryToSigmaRuleTool


class SigmaToolkit(BaseToolkit):
    """Sigma Toolkit."""

    sigmadb: VectorStore
    rule_creation_llm: BaseLanguageModel

    class Config:
        arbitrary_types_allowed = True

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
            FindSigmaRuleTool(sigmadb=self.sigmadb, llm=self.rule_creation_llm),
            TranslateSigmaRuleTool(),
            CreateSigmaRuleVectorStoreTool(sigmadb=self.sigmadb, llm=self.rule_creation_llm),
            QueryToSigmaRuleTool(llm=self.rule_creation_llm),
        ]



# stdlib
import asyncio
from typing import Type

# langchain
from langchain.prompts import ChatPromptTemplate
from langchain.pydantic_v1 import BaseModel, Field

# langchain typing
from langchain.schema.language_model import BaseLanguageModel
from langchain.schema.output_parser import StrOutputParser
from langchain.schema.runnable import RunnablePassthrough
from langchain.schema.vectorstore import VectorStore
from langchain.tools import BaseTool


class CreateSigmaRuleInput(BaseModel):
    """Input for TranslateSigmaRule tool, which uses SigmAIQ backend factory to convert a Sigma Rule into
    a query for a specific backend."""

    query: str = Field(
        description="The users question, used to search through the Sigma VectorStore and create a Sigma Rule."
    )

    class Config(BaseTool.Config):
        pass


class CreateSigmaRuleVectorStoreTool(BaseTool):
    """Class for translating Sigma rules via SigmAIQ Backend Factory"""

    name: str = "create_sigma_rule_vectorstore"
    args_schema: Type[BaseModel] = CreateSigmaRuleInput
    description: str = """Use this tool to take the users input, find similar Sigma Rules from the vectorstore,
then create a brand new Sigma Rule based on the users input and the similar Sigma Rules returned from the vectorstore
to use as context. The output is a Sigma Rule in YAML format.
"""
    sigmadb: VectorStore
    llm: BaseLanguageModel
    k: int = 3
    verbose = False
    verbose = False

    def _run(self, query: str) -> str:
        """Run the tool"""
        return asyncio.run(self._arun(query))

    async def _arun(self, query: str) -> str:
        """Async run the tool"""
        template = """You are a cybersecurity detection engineering assistant bot specializing in Sigma Rule creation.
You are assisting a user in creating a new Sigma Rule based on the users question.  
The users question is first used to find similar Sigma Rules from the a vectorstore containing official 
Sigma Rules. The official Sigma Rules can be used as context as needed in conjunction with the detection specified
in the users question to create a new Sigma Rule.  
The created Sigma Rule should be in YAML format and use the official Sigma schema.  The detection field
can contain multiple 'selection' identifiers and multiple 'filter' identifiers as needed, 
which can be used in the condition field to select criteria and filter out criteria respectively.
Set the 'author' to 'SigmAIQ (AttackIQ)', the date to today's date, and the reference to 'https://github.com/AttackIQ/SigmAIQ'.
If you use other rules as context and derive the created Sigma Rules from the context rules, you must
include the original authors under the 'author' field in the new rule in addition to "SigmAIQ (AttackIQ),
and add the original rule IDs under the 'related' field. The valid 'types' under 'related' are the following:

    derived: The rule was derived from the referred rule or rules, which may remain active.
    obsoletes: The rule obsoletes the referred rule or rules, which aren't used anymore.
    merged: The rule was merged from the referred rules. The rules may be still existing and in use.
    renamed: The rule had previously the referred identifier or identifiers but was renamed for whatever reason, e.g. from a private naming scheme to UUIDs, to resolve collisions etc. It's not expected that a rule with this id exists anymore.
    similar: Use to relate similar rules to each other (e.g. same detection content applied to different log sources, rule that is a modified version of another rule with a different level)

If you are unsure about the Sigma rule schema, you can get the information from the official
Sigma specification here first: https://raw.githubusercontent.com/SigmaHQ/sigma-specification/main/Sigma_specification.md

------------

Sigma Rule Schema:

title
id [optional]
related [optional]
   - id {{rule-id}}
     type {{type-identifier}}
status [optional]
description [optional]
references [optional]
author [optional]
date [optional]
modified [optional]
tags [optional]
logsource
   category [optional]
   product [optional]
   service [optional]
   definition [optional]
   ...
detection
   {{search-identifier}} [optional]
      {{string-list}} [optional]
      {{map-list}} [optional]
      {{field: value}}> [optional]
   ... # Multiple search identifiers can be specified as needed and used in the condition
   condition
fields [optional]
falsepositives [optional]
level [optional]:


------------

Vectorstore Search Results:

{context}

------------

User's Question: 
{query}
"""

        prompt = ChatPromptTemplate.from_template(template)
        retriever = self.sigmadb.as_retriever(search_kwargs={"k": self.k})
        chain = {"context": retriever, "query": RunnablePassthrough()} | prompt | self.llm | StrOutputParser()
        return await chain.ainvoke(query)



from sigma.backends.splunk import SplunkBackend
from sigma.backends.crowdstrike import LogScaleBackend
from sigmaiq.backends.sigmaiq_abstract_backend import AbstractGenericSigmAIQBackendClass


class SigmAIQCrowdstrikeSplunkBackend(AbstractGenericSigmAIQBackendClass, SplunkBackend):
    """SigmAIQ backend interface for the pySigma Splunk Backend library to translate a SigmaRule object
    to a Splunk search query with the Crowdstrike FDR format"""

    custom_formats = {}
    associated_pipelines = ["crowdstrike_fdr"]
    default_pipeline = "crowdstrike_fdr"
    formats = {"default": SplunkBackend.formats["default"]}


class SigmAIQCrowdstrikeLogscaleBackend(AbstractGenericSigmAIQBackendClass, LogScaleBackend):
    """SigmAIQ backend interface for the pySigma Logscale Backend library to translate a SigmaRule object
    to a Logscale search query with the Crowdstrike Falcon format"""

    custom_formats = {}
    associated_pipelines = ["crowdstrike_falcon"]
    default_pipeline = "crowdstrike_falcon"
    formats = {"default": LogScaleBackend.formats["default"]}



from sigma.backends.elasticsearch import LuceneBackend
from sigmaiq.backends.sigmaiq_abstract_backend import AbstractGenericSigmAIQBackendClass


class SigmAIQElasticsearchBackend(AbstractGenericSigmAIQBackendClass, LuceneBackend):
    """SigmAIQ backend interface for the pySigma Elasticsearch Backend library to translate a SigmaRule object
    to an Elasticsearch search query"""

    custom_formats = {}
    associated_pipelines = [
        "ecs_windows",
        "ecs_kubernetes",
        "ecs_windows_old",
        "ecs_zeek_beats",
        "ecs_zeek_corelight",
        "zeek_raw",
    ]
    default_pipeline = "ecs_windows"



from sigma.backends.kusto import KustoBackend
from sigmaiq.backends.sigmaiq_abstract_backend import AbstractGenericSigmAIQBackendClass


class SigmAIQDefenderXDRBackend(AbstractGenericSigmAIQBackendClass, KustoBackend):
    """SigmAIQ backend interface for the pySigma Kusto Backend library to translate a SigmaRule object
    to a Kusto search query with the Microsoft Defender XDR format"""

    custom_formats = {}
    associated_pipelines = ["microsoft_xdr"]
    default_pipeline = "microsoft_xdr"


class SigmAIQSentinelASIMBackend(AbstractGenericSigmAIQBackendClass, KustoBackend):
    """SigmAIQ backend interface for the pySigma Kusto Backend library to translate a SigmaRule object
    to a Kusto search query with the Microsoft Sentinel ASIM format"""

    custom_formats = {}
    associated_pipelines = ["sentinel_asim"]
    default_pipeline = "sentinel_asim"


class SigmAIQAzureMonitorBackend(AbstractGenericSigmAIQBackendClass, KustoBackend):
    """SigmAIQ backend interface for the pySigma Kusto Backend library to translate a SigmaRule object
    to a Kusto search query with the Microsoft Azure Monitor format"""

    custom_formats = {}
    associated_pipelines = ["azure_monitor"]
    default_pipeline = "azure_monitor"



from sigma.backends.splunk import SplunkBackend
from sigma.rule import SigmaRule
from sigma.collection import SigmaCollection
from sigmaiq.backends.sigmaiq_abstract_backend import AbstractGenericSigmAIQBackendClass
from copy import copy, deepcopy
import re
from importlib_resources import files

SAVEDSEARCHES_TEMPLATE = files("sigmaiq.backends.splunk").joinpath("savedsearches_template.txt").read_text()


class SigmAIQSplunkBackend(AbstractGenericSigmAIQBackendClass, SplunkBackend):
    """SigmAIQ backend interface for the pySigma Splunk Backend library to translate a SigmaRule object
    to a Splunk search query"""

    custom_formats = {"stanza": "Enterprise Security savedsearches.conf stanza"}
    associated_pipelines = ["splunk_windows", "splunk_wineventlog", "splunk_windows_sysmon_acc", "splunk_cim_dm"]
    default_pipeline = "splunk_windows"

    def handle_output_format(self, sigma_rule, output):
        """Converts Splunk search to savedsearches.conf stanza if output_format given"""
        if self.custom_output_format == "stanza":
            output = self._convert_to_stanza(sigma_rule, output)
        return output

    def _convert_to_stanza(self, sigma_rule, output: list):
        """Converts a list of SigmaRule Splunk query outputs to savedsearches.conf stanzas to use in
        Splunk Enterprise Security Correlation Searches
        """
        # Convert SigmaRule to SigmaCollection just for ease of use
        sigma_rule_temp = copy(sigma_rule)
        if isinstance(sigma_rule_temp, SigmaRule):
            collection = SigmaCollection([sigma_rule_temp])
        else:
            collection = sigma_rule_temp

        stanzas = []
        for i, rule in enumerate(collection):
            title = rule.title
            search = output[i]
            tags = self._extract_mitre_tags(rule.tags)
            replacements = {
                "%TITLE_PLACEHOLDER%": title,
                "%MITRE_ATTACK_PLACEHOLDER%": f"{tags}",
                "%NOTABLE_TITLE_PLACEHOLDER%": title,
                "%DRILLDOWN_TITLE_PLACEHOLDER%": title,
                "%DRILLDOWN_SEARCH_PLACEHOLDER%": search,
                "%SEARCH_PLACEHOLDER%": search,
            }

            stanza = deepcopy(SAVEDSEARCHES_TEMPLATE)
            for k, v in replacements.items():
                stanza = stanza.replace(k, v)
            stanzas.append(stanza)
        return stanzas

    @staticmethod
    def _extract_mitre_tags(tags):
        """Takes tags from a SigmaRule and outputs them into a format that a Splunk Correlation Search stanza will
        recognize for the action.correlationsearch.annotations field:
        i.e.    action.correlationsearch.annotations = {"mitre_attack":["T1055.011","T1560.001","T1205.002","T1583"]}
        """
        mitre_tags = {"mitre_attack": []}
        if not tags:
            return mitre_tags
        for tag in tags:
            matches = re.match(r"^(attack\.)(t\d{4}(\.\d{3})?)", str(tag))
            if matches:
                mitre_tags["mitre_attack"].append(matches[2].upper())
        return mitre_tags



from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder

# Default prompts for Sigma agents
SIGMA_AGENT_PROMPT = prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are a threat detection engineering assistant bot specializing in Sigma Rules."
            "You have four tools at your disposal: "
            "1. translate_sigma_rule: converts or translates a Sigma Rule into a query for a specific backend/product."
            "2. find_sigma_rule: Searches for a Sigma Rule in the vector database based on the users question."
            "3. create_sigma_rule_vectorstore: Creates new Sigma Rule from the users input, as well as rules in a sigma rule vectorstore to use as context based on the users question. If the user's question already contains a query, use 'query_to_sigma_rule' instead."
            "4. query_to_sigma_rule: Converts/translates a product/SIEM/backend query or search from the query language into a YAML Sigma Rule."
            "Do not use 'translate_sigma_rule' unless the user explicitly asks for a Sigma Rule to be converted or translated "
            "into a query for a specific backend, pipeline, and/or output format.",
        ),
        ("user", "{input}"),
        MessagesPlaceholder(variable_name="agent_scratchpad"),
    ]
)



from datetime import datetime
from typing import Union

from sigma.collection import SigmaCollection
from sigma.rule import SigmaRule


def _is_v1_schema(rule_data: dict) -> bool:
    """Check if the rule uses v1 schema patterns."""
    if not isinstance(rule_data, dict):
        return False

    # Check date format
    date_str = rule_data.get("date")
    if date_str and "/" in date_str:
        return True

    # Check modified format
    modified_str = rule_data.get("modified")
    if modified_str and "/" in modified_str:
        return True

    # Check tags format
    tags = rule_data.get("tags", [])
    for tag in tags:
        if any(ns in tag for ns in ["attack-", "attack_", "cve-", "detection-"]):
            return True

    # Check related field
    related = rule_data.get("related", [])
    for rel in related:
        if rel.get("type") == "obsoletes":
            return True

    return False


def _convert_to_v2_schema(rule_data: dict) -> dict:
    """Convert v1 schema rule to v2 schema."""
    rule_data = rule_data.copy()

    # Convert date and modified format
    if "date" in rule_data and "/" in rule_data["date"]:
        try:
            date_obj = datetime.strptime(rule_data["date"], "%Y/%m/%d")
            rule_data["date"] = date_obj.strftime("%Y-%m-%d")
        except ValueError:
            pass

    if "modified" in rule_data and "/" in rule_data["modified"]:
        try:
            date_obj = datetime.strptime(rule_data["modified"], "%Y/%m/%d")
            rule_data["modified"] = date_obj.strftime("%Y-%m-%d")
        except ValueError:
            pass

    # Convert tags
    if "tags" in rule_data:
        new_tags = []
        for tag in rule_data["tags"]:
            # Convert common namespace patterns
            tag = tag.replace("attack-", "attack.")
            tag = tag.replace("attack_", "attack.")
            tag = tag.replace("cve-", "cve.")
            tag = tag.replace("detection-", "detection.")
            new_tags.append(tag)
        rule_data["tags"] = new_tags

    # Convert related field
    if "related" in rule_data:
        for rel in rule_data["related"]:
            if rel.get("type") == "obsoletes":
                rel["type"] = "obsolete"

    return rule_data


def create_sigma_rule_obj(sigma_rule: Union[SigmaRule, SigmaCollection, dict, str, list]):
    """Checks sigma_rule to ensure it's a SigmaRule or SigmaCollection object. It can also be a valid Sigma rule
    representation in a dict or yaml str (or list of valid dicts/yaml strs) that can be used with SigmaRule class methods to
    create a valid SigmaRule object. The following checks are performed:
        - If the object is a SigmaRule or SigmaCollection object, return the object.
        - If the object is a list, we will recursively check each element in the list and return a SigmaCollection
        of the returned objects
        - If the object is a dict, use SigmaRule.from_dict(sigma_rule) to create a new SigmaRule object.
        - If the object is a str, use SigmaRule.from_yaml(sigma_rule) to create a new SigmaRule object.

    :param sigma_rule: A list of, or single SigmaRule, SigmaCollection, or valid dict/yaml str that will be
    returned as a SigmaRule or SigmaCollection object.
    :return: SigmaRule or SigmaCollection object as-is or created using SigmaRule classmethods.
    """

    if isinstance(sigma_rule, (SigmaRule, SigmaCollection)):
        return sigma_rule
    if isinstance(sigma_rule, list):
        rules = []
        for s in sigma_rule:
            result = create_sigma_rule_obj(s)
            if isinstance(result, SigmaCollection):
                rules.extend(result.rules)
            else:
                rules.append(result)
        return SigmaCollection(rules)
    if isinstance(sigma_rule, dict):
        # Check and convert v1 schema if needed
        if _is_v1_schema(sigma_rule):

            sigma_rule = _convert_to_v2_schema(sigma_rule)
        return SigmaRule.from_dict(sigma_rule)
    if isinstance(sigma_rule, str):
        # For YAML strings, we need to parse to dict first
        try:
            import yaml

            rule_dict = yaml.safe_load(sigma_rule)
            if _is_v1_schema(rule_dict):
                rule_dict = _convert_to_v2_schema(rule_dict)
            return SigmaRule.from_dict(rule_dict)
        except Exception as e:
            print(e)
            return SigmaRule.from_yaml(sigma_rule)
    raise TypeError(
        f"Invalid type '{type(sigma_rule)}' for `sigma_rule`. "
        f"Use a SigmaRule, SigmaCollection, dict, str, or list of these types instead."
    )



# stdlib
import os
from typing import Type, List

# sigmaiq
from sigmaiq.utils.sigma.rule_updater import SigmaRuleUpdater
from sigmaiq.globals import DEFAULT_DIRS

# langchain
from langchain_openai import OpenAIEmbeddings
from langchain_community.vectorstores import FAISS
from langchain_community.document_loaders import DirectoryLoader, TextLoader
from langchain.text_splitter import CharacterTextSplitter

# langchain typing
from langchain.schema.embeddings import Embeddings
from langchain.schema.vectorstore import VectorStore
from langchain.docstore.document import Document
from langchain.document_loaders.base import BaseLoader
from langchain.schema.document import BaseDocumentTransformer


class SigmaLLM(SigmaRuleUpdater):
    """Base class for Sigma rules with LLMs.
    Provides methods for ensuring the latest Sigma rule package is installed, creating embeddings from Sigma rules,
    and storing them in a vector store. Also provides basic search functionality based on the vector store embeddings.

    All agents, tools, and toolkits are in separate classes.

    To use custom embeddings, text splitters, loaders, etc, provide them as args override the methods in this class as
    needed by your custom passed classes.
    """

    def __init__(
        self,
        rule_dir: str = None,
        vector_store_dir: str = None,
        embedding_model: OpenAIEmbeddings = None,
        embedding_function: Type[Embeddings] = OpenAIEmbeddings,  # TODO RS : Consolidate this with embedding_model
        vector_store: Type[VectorStore] = FAISS,
        rule_loader: Type[BaseLoader] = DirectoryLoader,
        rule_splitter: Type[BaseDocumentTransformer] = CharacterTextSplitter,
    ):
        """Initializes the SigmaLLM object.
        If passing custom embeddings, vector stores, loaders, or splitters, pass the class itself rather than
        an instance of the class. For example, pass `OpenAIEmbeddings` instead of `OpenAIEmbeddings()`. Then, override
        the methods in this class to load the custom classes if needed.

        Requires environmental variable `OPENAI_API_KEY` to be set to your OpenAI API key if using any OpenAI models
        or embeddings.

        Args:
            rule_dir (str, optional): The directory to store the Sigma rules. Defaults to None.
            vector_store_dir (str, optional): The directory to store the vector store. Defaults to None.
            embedding_function (Type[Embeddings], optional): The Embeddings class to use for Sigma rule embeddings. Defaults to OpenAIEmbeddings.
            vector_store (Type[VectorStore], optional): The VectorStore class to use for Sigma rule embeddings. Defaults to FAISS.
            rule_loader (Type[BaseLoader], optional): The DocumentLoader class to use for loading Sigma rules. Defaults to DirectoryLoader.
            rule_splitter (Type[BaseDocumentTransformer], optional): The DocumentTransformer class to use for splitting Sigma rules. Defaults to CharacterTextSplitter.
        """
        # Download/update sigma rules from parent class
        super().__init__(rule_dir=rule_dir)

        # Setup rest of class
        self.vector_store_dir = self._setup_vector_store_dir(vector_store_dir)
        if embedding_model:
            self.embedding_function = embedding_model
        else:
            self.embedding_function = embedding_function()
        self.vector_store = vector_store
        self.sigmadb = None
        self.rule_loader = rule_loader
        self.rule_splitter = rule_splitter

    def load_sigma_vectordb(self):
        """Loads the Sigma rule vector store.
        Override `load_local()` below with how your vector store class loads local Vector DBs"""
        if not os.path.exists(self.vector_store_dir):
            raise FileNotFoundError(f"VectorStore not found at {self.vector_store_dir}.")
        try:
            self.sigmadb = self.vector_store.load_local(  # CHANGE ME IF NEEDED
                folder_path=self.vector_store_dir,
                embeddings=self.embedding_function,
                allow_dangerous_deserialization=True,
            )
        except Exception as e:
            raise e

    def create_sigma_vectordb(self, save: bool = True):
        """Creates Sigma rule vector store by performing the following actions:
            1. Load each Sigma rule from the local SigmaHQ Sigma rules repository as Documents
            2. Split each Sigma rule Document
            3. Embed each Sigma rule Document and store in VectorStore
            4. Save the vectordb (if arg set) to disk

        Each of these steps has its own associated method; override them if you would like to change its behavior, for
        example, by using a different TextSplitter or VectorStore.

        Args:
            save (bool, optional): If True, will save the VectorStore to disk. Defaults to True.

        """
        if not self.installed_tag:
            self.update_sigma_rules()

        # Load Sigma docs
        sigma_docs = self.create_sigma_rule_docs()
        print(f"Loaded {len(sigma_docs)} Sigma rules")
        # Split Sigma docs
        sigma_docs = self.split_sigma_docs(sigma_docs)
        # Create VectorStore
        self.create_vectordb(sigma_docs)
        print(f"Created Sigma vectordb at {self.vector_store_dir}")
        # Save VectorStore
        if save:
            self.save_vectordb()

    def create_sigma_rule_docs(self) -> List[Document]:
        """Generator to loads Sigma rules from the local SigmaHQ Sigma rules repository."""
        sigma_rule_docs = []
        sigma_rule_docs += self.rule_loader(self.rule_dir, glob="**/*.yml", loader_cls=TextLoader).load()

        return sigma_rule_docs

    def split_sigma_docs(self, sigma_docs) -> List[Document]:
        """Splits Sigma rule Documents into chunks based on the DirectoryLoader provided on initialization.
        By default, we don't want to split up rules much, as we want the whole rule embedded. If you want to split
        rules into smaller chunks, override this method and return the chunks, or use your own text splitter in initialization.

        Args:
            sigma_docs (List[Document]): The list of Sigma rule Documents to split.

        Returns:
            List[Document]: The list of Sigma rule Documents, split into chunks.
        """
        # Override if needed
        self.rule_splitter = self.rule_splitter(chunk_size=99999)  # only chunk if rule is larger than 99999 characters
        return self.rule_splitter.split_documents(sigma_docs)

    def create_vectordb(self, sigma_docs: List[Document]):
        """Creates the VectorStore from the Sigma rule Documents.
        Override `from_documents()` below with how your vector store class adds documents to the Vector DB"""
        self.sigmadb = self.vector_store.from_documents(sigma_docs, self.embedding_function)  # CHANGE ME IF NEEDED

    def save_vectordb(self, vectordb_path: str = None):
        """Saves the VectorStore to disk. If no path is provided, will save to the path provided on initialization.
        Override `save_local()` below with how your vector store class saves local Vector DBs

        Args:
            vectordb_path (str, optional): The path to save the VectorStore. Defaults to None and will use the path provided on initialization.
        """

        self.sigmadb.save_local(self.vector_store_dir)

    @staticmethod
    def _setup_vector_store_dir(vector_store_dir: str = None) -> str:
        """Checks if the vector store directory exists. If not, creates it.

        Args:
            vector_store_dir (str, optional): The directory to store the vector store. Defaults to None.

        Returns:
            str: The vector store directory path.
        """
        if not vector_store_dir:
            vector_store_dir = DEFAULT_DIRS.VECTOR_STORE_DIR
        if not os.path.exists(vector_store_dir):
            os.makedirs(vector_store_dir)
        return vector_store_dir

    def simple_search(self, query: str, k: int = 3) -> List[Document]:
        """Searches the Sigma rule vector store for the query text using similarity search.

        Args:
            query (str): The query text to search for.
            k (int, optional): The number of results to return. Defaults to 3.

        Returns:
            List[Document]: The top 'k' matching Sigma Rules from the search.
        """
        if not self.sigmadb:
            self.load_sigma_vectordb()
        return self.sigmadb.similarity_search(query, k)



from sigma.backends.stix import stixBackend
from sigmaiq.backends.sigmaiq_abstract_backend import AbstractGenericSigmAIQBackendClass


class SigmAIQStixBackend(AbstractGenericSigmAIQBackendClass, stixBackend):
    associated_pipelines = ["stix_2_0", "stix_shifter"]
    default_pipeline = "stix_2_0"

    # Remove "stix" format until its not broken :(
    formats = {"default": stixBackend.formats["default"]}



import logging
from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

from sigma.collection import SigmaCollection
from sigma.processing.pipeline import ProcessingPipeline
from sigma.rule import SigmaRule

# Backends
from sigmaiq.backends.carbonblack import SigmAIQCarbonBlackBackend
from sigmaiq.backends.cortexxdr import SigmAIQCortexXDRBackend
from sigmaiq.backends.crowdstrike import (
    SigmAIQCrowdstrikeLogscaleBackend,
    SigmAIQCrowdstrikeSplunkBackend,
)
from sigmaiq.backends.elasticsearch import SigmAIQElasticsearchBackend
from sigmaiq.backends.insightidr import SigmAIQInsightIDRBackend
from sigmaiq.backends.kusto import SigmAIQAzureMonitorBackend, SigmAIQDefenderXDRBackend, SigmAIQSentinelASIMBackend
from sigmaiq.backends.netwitness import SigmAIQNetwitnessBackend
from sigmaiq.backends.opensearch import SigmAIQOpensearchBackend
from sigmaiq.backends.qradar import SigmAIQQRadarBackend
from sigmaiq.backends.secops import SigmAIQSecOpsBackend
from sigmaiq.backends.sentinelone import SigmAIQSentinelOneBackend
from sigmaiq.backends.sigma import SigmAIQSigmaBackend

## Abstract
from sigmaiq.backends.sigmaiq_abstract_backend import AbstractGenericSigmAIQBackendClass
from sigmaiq.backends.splunk import SigmAIQSplunkBackend
from sigmaiq.backends.stix import SigmAIQStixBackend
from sigmaiq.exceptions import InvalidSigmAIQBackend
from sigmaiq.sigmaiq_pipeline_factory import SigmAIQPipeline, SigmAIQPipelineResolver

# Utils
from sigmaiq.utils.sigmaiq.sigmaiq_utils import create_sigma_rule_obj

AVAILABLE_BACKENDS = {
    "carbonblack": "Carbon Black EDR",
    "cortexxdr": "Palo Alto Cortex XDR",
    "crowdstrike_splunk": "Crowdstrike FDR Splunk Query",
    "crowdstrike_logscale": "Crowdstrike Logscale Query",
    "elasticsearch": "Elastic Elasticsearch SIEM",
    "insightidr": "Rapid7 InsightIDR SIEM",
    "microsoft_xdr": "Microsoft XDR Advanced Hunting Query (KQL) (Defender, Office365, etc)",
    "microsoft_sentinel_asim": "Microsoft Sentinel ASIM Query (KQL)",
    "microsoft_azure_monitor": "Microsoft Azure Monitor Query (KQL)",
    "netwitness": "Netwitness Query",
    "opensearch": "OpenSearch Lucene",
    "qradar": "IBM QRadar",
    "secops": "Google SecOps (Chronicle)",
    "sentinelone": "SentinelOne EDR",
    "splunk": "Splunk SIEM",
    "sigma": "Original YAML/JSON Sigma Rule Output",
    "stix": "STIX 2.0 & STIX Shifter Queries",
}


class SigmAIQBackend:
    """
    Implement the Factory pattern for the pySigma Backends, which provide the translation from SigmaRule objects
    to SIEM/Security Tool queries.
    """

    def __init__(
        self,
        backend: str,
        processing_pipeline: Optional[Union[str, list, ProcessingPipeline]] = None,
        output_format: Optional[str] = None,
    ):
        """Initialize instance attributes.

        :param backend: Specifies the desired backend.
        :type backend: str
        :param processing_pipeline: Submit processing pipeline to pySigma backend to automatically apply processing pipeline
        when converting a SigmaRule object to the backend query. This can also be done manually outside the backend,
        defaults to None
        :type processing_pipeline: ProcessingPipeline, optional
        :param output_format: Format specific to each SigmAIQBackendClass to output specific formats, such as
        'yaml' or 'json' for the raw Sigma Backend, defaults to None
        :type output_format: str, optional
        """
        logging.debug(f"Executing SigmAIQBackend constructor. backend: {backend}")
        self.backend = backend
        self.processing_pipeline = self._setup_processing_pipeline(processing_pipeline)
        self.output_format = output_format

    def create_backend(self) -> AbstractGenericSigmAIQBackendClass:
        """
        Create the factory agents.

        :return: The custom AbstractGenericSigmAIQBackendClass instance of the pySigma Backend depending on the
        value of the `backend` parameter. The abstract backend class inherits the TextBackend and specific pySigma
        backend classes attributes and methods.
        :rtype: AbstractGenericSigmAIQBackendClass
        """
        kwargs = {"processing_pipeline": self.processing_pipeline, "output_format": self.output_format}

        # Carbon Black EDR (standard & enterprise)
        if self.backend == "carbonblack":
            return SigmAIQCarbonBlackBackend(**kwargs)
        # Cortex XDR, Palo Alto
        if self.backend == "cortexxdr":
            return SigmAIQCortexXDRBackend(**kwargs)
        # Crowdstrike Splunk Query
        if self.backend == "crowdstrike_splunk":
            pipelines = ["crowdstrike_fdr", kwargs["processing_pipeline"]]
            kwargs["processing_pipeline"] = SigmAIQPipelineResolver(pipelines).process_pipelines()
            return SigmAIQCrowdstrikeSplunkBackend(**kwargs)
        # Crowdstrike Logscale Query
        if self.backend == "crowdstrike_logscale":
            pipelines = ["crowdstrike_falcon", kwargs["processing_pipeline"]]
            kwargs["processing_pipeline"] = SigmAIQPipelineResolver(pipelines).process_pipelines()
            return SigmAIQCrowdstrikeLogscaleBackend(**kwargs)
        # Elasticsearch
        if self.backend == "elasticsearch":
            return SigmAIQElasticsearchBackend(**kwargs)
        # InsightIDR
        if self.backend == "insightidr":
            return SigmAIQInsightIDRBackend(**kwargs)
        # Microsoft Kusto
        if self.backend == "microsoft_xdr":
            return SigmAIQDefenderXDRBackend(**kwargs)
        if self.backend == "microsoft_sentinel_asim":
            return SigmAIQSentinelASIMBackend(**kwargs)
        if self.backend == "microsoft_azure_monitor":
            return SigmAIQAzureMonitorBackend(**kwargs)
        # Netwitness
        if self.backend == "netwitness":
            return SigmAIQNetwitnessBackend(**kwargs)
        # Opensearch
        if self.backend == "opensearch":
            return SigmAIQOpensearchBackend(**kwargs)
        #
        # QRadar Backend
        if self.backend == "qradar":
            return SigmAIQQRadarBackend(**kwargs)
        # SecOps Backend
        if self.backend == "secops":
            return SigmAIQSecOpsBackend(**kwargs)
        # SentinelOne
        if self.backend == "sentinelone":
            return SigmAIQSentinelOneBackend(**kwargs)
        # Splunk Backend
        if self.backend == "splunk":
            if kwargs["output_format"] == "data_model":
                pipelines = [p for p in ["splunk_cim_dm", kwargs.get("processing_pipeline")] if p is not None]
                kwargs["processing_pipeline"] = SigmAIQPipelineResolver(pipelines).process_pipelines()
            return SigmAIQSplunkBackend(**kwargs)
        # Raw sigma output
        if self.backend == "sigma":
            if not kwargs.get("output_format"):
                kwargs["output_format"] = "yaml"
            return SigmAIQSigmaBackend(**kwargs)
        # STIX
        if self.backend == "stix":
            return SigmAIQStixBackend(**kwargs)

        raise InvalidSigmAIQBackend(
            'Backend not supported: "{}". Available backends:\n{}'.format(
                self.backend, "\n".join([f"{k}: {v}" for k, v in AVAILABLE_BACKENDS.items()])
            )
        )

    @staticmethod
    def _setup_processing_pipeline(processing_pipeline):
        """Return the processing pipeline if None or already a ProcessingPipeline
        Otherwise, try to create them with our processing_pipeline_factory"""

        if processing_pipeline:
            if not isinstance(processing_pipeline, list):
                processing_pipeline = [processing_pipeline]
        else:
            processing_pipeline = []

        # Use pipeline resolver to combine pipelines
        if processing_pipeline:
            return SigmAIQPipelineResolver(processing_pipelines=processing_pipeline).process_pipelines()
        return None

    @classmethod
    def create_all_and_translate(
        cls,
        sigma_rule: Union[SigmaRule, SigmaCollection, str, dict],
        show_errors: Optional[bool] = False,
        excluded_backends: Optional[List[str]] = None,
    ) -> Dict[Any, Any]:
        """Iterates through all combinations of backends, associated pipelines with each backend, and output formats
        for each backend, and creates a dict of outputs.

        :param sigma_rule: A valid SigmaRule or SigmaCollection object to translate
        :type sigma_rule: Union[SigmaRule, SigmaCollection]
        :param show_errors: If True, errors will be included in the list of outputs. Errors can include errors from a
        backend when specific fields cannot be converted to a query. Defaults to False
        :type show_errors: bool
        :param excluded_backends: List of backends to exclude from translations. Defaults to None
        :return: Dict of output results in the following format:
            {backend: {pipeline: {output_format: [queries]}
        :rtype: Dict[Any, Any]
        """
        backends_pipelines = cls.display_all_associated_pipelines()
        backends_output_formats = cls.display_backends_and_outputs()
        excluded_backends = [x.lower() for x in excluded_backends] if excluded_backends else []
        results = {}
        sigma_rule = create_sigma_rule_obj(sigma_rule)
        for backend, pipelines in backends_pipelines.items():
            if backend.lower() in excluded_backends:
                continue
            backend_obj = cls(backend=backend).create_backend()
            for pipeline in pipelines:
                backend_obj.set_pipeline(pipeline)
                for output_format in backends_output_formats[backend].get("output_formats"):
                    backend_obj.set_output_format(output_format)
                    output = []
                    try:
                        output = backend_obj.translate(deepcopy(sigma_rule))
                    except Exception as exc:
                        if show_errors:
                            output = [exc]
                    if output:
                        if backend not in results.keys():
                            results[backend] = {}
                        if pipeline not in results[backend].keys():
                            results[backend][pipeline] = {}
                        results[backend][pipeline][output_format] = output
        return results

    @classmethod
    def display_available_backends(cls) -> Dict[str, str]:
        """Simple class method for returning dict of available backends and descriptions
        :return: Dict containing backend name and description
        :rtype: Dict[str, str]
        """
        return AVAILABLE_BACKENDS

    @classmethod
    def display_backends_and_outputs(cls) -> Dict:
        """Display all available backends as well as their associated output formats, both standard and SigmAIQ
        custom

        :return: Dict containing available backends for keys, and a dict of output formats with descriptions as values
        :rtype: Dict
        """
        backend_formats = {}
        for backend, description in AVAILABLE_BACKENDS.items():
            backend_instance = cls(backend=backend).create_backend()
            output_formats = backend_instance.formats or {"default": "Default output format"}
            custom_output_formats = backend_instance.custom_formats or {}
            output_formats = {**output_formats, **custom_output_formats} if custom_output_formats else output_formats
            backend_formats[backend] = {}
            backend_formats[backend]["description"] = description
            backend_formats[backend]["output_formats"] = output_formats
        return backend_formats

    @classmethod
    def display_all_associated_pipelines(cls) -> Dict:
        """Displays all available backends with their associated pipelines. Output format is a dict with the available
        pipeline as a key and a dict of the pipeline name and descriptions as the value.

        :return: Dict of backend pipline information
        :rtype: Dict

        """
        associated_pipelines = {}
        available_pipelines = SigmAIQPipeline.display_available_pipelines()
        for backend in AVAILABLE_BACKENDS.keys():
            pipeline_desc = {}
            backend_instance = cls(backend=backend).create_backend()
            pipelines = backend_instance.associated_pipelines
            for pipeline in pipelines:
                pipeline_desc[pipeline] = available_pipelines[pipeline]
            associated_pipelines[backend] = pipeline_desc

        return associated_pipelines



import asyncio
import json
from typing import Union, Type

from langchain.tools import BaseTool
from pydantic import BaseModel, Field, Extra

from sigmaiq.sigmaiq_backend_factory import AVAILABLE_BACKENDS, SigmAIQBackend
from sigmaiq.sigmaiq_pipeline_factory import AVAILABLE_PIPELINES


class TranslateSigmaRuleInput(BaseModel):
    """Input for TranslateSigmaRule tool, which uses SigmAIQ backend factory to convert a Sigma Rule into
    a query for a specific backend."""

    sigma_rule: Union[str, dict] = Field(
        default=None,
        description="""The Sigma Rule to translate. This can be one of two formats:

1. A YAML string of the Sigma rule, with at least the title, logsource, and detection fields.
2. A dict object of the Sigma rule, which is the same as the YAML string and must contain the 
"title, logsource, and detection fields""",
    )
    backend: str = Field(
        default=None,
        description="""The backend or product to translate the Sigma rule to. Backend options their descriptions are as 
follows:\n"""
        + f"{json.dumps(AVAILABLE_BACKENDS, indent=2)}",
    )
    processing_pipeline: str = Field(
        default=None,
        description="""The processing pipeline to use for the Sigma rule. This should only be set if explicitly provided
by the user, as certain pipelines are only compatible with certain backends. Otherwise, set this to None.
Pipeline options and their 
descriptions are as follows:\n"""
        + f"{json.dumps({k: v['description'] for k, v in AVAILABLE_PIPELINES.items()}, indent=2)}",
    )
    output_format: str = Field(
        default="default",
        description="""The output format for the translated rule. Unless specified, 'default' should be used, as this is
    the option available in all backends. Each backend option and valid backends with their
    descriptions are as follows:\n"""
        + f"{json.dumps({k: v['output_formats'] for k, v in SigmAIQBackend.display_backends_and_outputs().items()}, indent=2)}",
    )


class TranslateSigmaRuleTool(BaseTool):
    """Class for translating Sigma rules via SigmAIQ Backend Factory"""

    name: str = "translate_sigma_rule"
    args_schema: Type[BaseModel] = TranslateSigmaRuleInput
    description: str = """
Use this tool to translate or convert a Sigma rule into a query for a specific backend.
The input must be a Sigma Rule, which can be provided as a YAML string or dict object.
Additionally, the backend (product) must be specified, and the processing pipeline and output format can be 
optionally specified.
The output is json of the translated rule to a query for the backend, or an error message if the 
translation fails.
"""
    # return_direct = True  # We don't need an agent LLM to think about the output, it is what it is.
    verbose = False

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    def _run(
        self,
        sigma_rule: Union[str, dict] = None,
        backend: str = None,
        processing_pipeline: str = None,
        output_format: str = "default",
    ) -> str:
        """Run the tool"""
        return asyncio.run(self._arun(sigma_rule, backend, processing_pipeline, output_format))

    async def _arun(
        self,
        sigma_rule: Union[str, dict] = None,
        backend: str = None,
        processing_pipeline: str = None,
        output_format: str = "default",
    ) -> str:
        """Async run the tool"""
        # Get backend object
        backend_obj = SigmAIQBackend(
            backend=backend, processing_pipeline=processing_pipeline, output_format=output_format
        ).create_backend()

        try:
            output = backend_obj.translate(sigma_rule)[0]
        except Exception as e:
            output = f"ERROR: {e}"
        # Return translated rule
        return output



from sigma.backends.QRadarAQL import QRadarAQLBackend
from sigmaiq.backends.sigmaiq_abstract_backend import AbstractGenericSigmAIQBackendClass


class SigmAIQQRadarBackend(AbstractGenericSigmAIQBackendClass, QRadarAQLBackend):
    """SigmAIQ backend interface for the pySigma QRadar-AQL Backend library to translate a SigmaRule object
    to a Splunk search query"""

    associated_pipelines = ["qradar_fields", "qradar_payload"]
    default_pipeline = "qradar_fields"



from sigma.pipelines.splunk import splunk_windows_pipeline
from sigma.pipelines.windows import windows_audit_pipeline
from sigma.processing.resolver import ProcessingPipelineResolver
from sigma.processing.pipeline import ProcessingPipeline


def splunk_wineventlog_pipeline() -> ProcessingPipeline:
    """Returns ProcessingPipeline that combines the windows_audit pipeline with the Splunk Windows pipeline
    for using Windows Event Log field names in Splunk searches rather than Sysmon.
    """
    pipelines = [windows_audit_pipeline(), splunk_windows_pipeline()]
    resolver = ProcessingPipelineResolver.from_pipeline_list(pipelines)
    pipeline = resolver.resolve(resolver.pipelines)
    pipeline.name = "Splunk WinEventLog Pipeline"
    return pipeline


splunk_wineventlog_pipeline()



import os
import shutil

import requests
from sigmaiq.globals import DEFAULT_DIRS
from pathlib import Path
import zipfile
import io


class SigmaRuleUpdater:
    """Download/update Sigma rules from the official SigmaHQ release packages."""

    PACKAGE_NAME_URIS = {
        "core": "sigma_core.zip",
        "core+": "sigma_core+.zip",
        "core++": "sigma_core++.zip",
        "emerging_threats": "sigma_emerging_threats_addon.zip",
        "all": "sigma_all_rules.zip",
    }

    BASE_DOWNLOAD_URL = "https://github.com/SigmaHQ/sigma/releases/download"

    def __init__(self, rule_dir: str = None):
        self.rule_dir = setup_rule_dir(rule_dir)
        self.installed_tag = self._get_installed_release_tag()
        self.latest_tag = get_latest_sigma_release_tag()

    def _get_installed_release_tag(self) -> str:
        """Returns the currently installed Sigma release tag by checking the directory name of the
        SIGMA_RULE_DIR directory.

        Returns:
            str: The currently installed Sigma release tag, or "" if not found.
        """
        version_file = Path(self.rule_dir) / "version.txt"
        if not version_file.exists():
            return ""
        with open(version_file, "r") as f:
            current_version = "r" + f.readlines()[0].split(": ")[-1].strip()
        return current_version

    def _needs_update(self) -> bool:
        """Checks if the currently installed Sigma release tag is the same as the latest release tag.

        Returns:
            bool: True if the latest_tag is different from the currently installed release tag, False otherwise.
        """
        if not self.installed_tag:
            return True
        return self.latest_tag != self.installed_tag

    def update_sigma_rules(
        self,
        force: bool = False,
        package_name: str = "core",
        emerging_threats: bool = False,
    ):
        """Downloads the latest Sigma release package, if needed, and extracts it to the SIGMA_RULE_DIR directory.

        Args:
            force (bool, optional): If True, will always download the latest Sigma release package. Defaults to False.
            package_name (str, optional): The name of the Sigma release package to download. Defaults to "core". Valid
            options are "core", "core+", "core++", "emerging_threats", and "all"
            emerging_threats (bool, optional): If True, will download the emerging_threats Sigma release package in
            addition to the package specified in args. Defaults to False.

        Raises:
            ValueError: If the package_name is invalid.


        """
        if package_name not in self.PACKAGE_NAME_URIS.keys():
            raise ValueError(f"Invalid package name '{package_name}'. Valid options are: {self.PACKAGE_NAME_URIS.keys}")

        print(f"Installed Sigma release tag at {self.rule_dir}: {self.installed_tag}")
        if not self._needs_update() and not force:
            print("Sigma rules are up-to-date.")
            return
        print(f"Updating Sigma rules to {self.latest_tag}...")
        self._download_sigma_release(package_name)

        if emerging_threats:
            if package_name in ["emerging_threats", "all"]:
                print("emerging-threats already contains in the selected package, skipping download.")
            else:
                print("Downloading emerging_threats Sigma rules...")
                self._download_sigma_release("emerging_threats", overwrite=False)
        print("Sigma rules up to date!")

    def _download_sigma_release(self, package_name: str, overwrite: bool = True):
        """Downloads the latest Sigma release package and extracts it to the SIGMA_RULE_DIR directory.

        Args:
            package_name (str): The name of the Sigma release package to download. Valid options are "core", "core+",
            "core++", "emerging_threats", and "all"
            overwrite (bool, optional): If True, will overwrite the currently installed Sigma release package. Defaults
            to True.

        Raises:
            ValueError: If the package_name is invalid.
        """
        if overwrite:
            print(f"Removing Sigma rules at {self.rule_dir} before new download...")
            shutil.rmtree(os.path.join(self.rule_dir, "rules"), ignore_errors=True)
            shutil.rmtree(os.path.join(self.rule_dir, "rules-emerging-threats"), ignore_errors=True)

        if package_name not in self.PACKAGE_NAME_URIS.keys():
            raise ValueError(f"Invalid package name '{package_name}'. Valid options are: {self.PACKAGE_NAME_URIS.keys}")

        url = f"{self.BASE_DOWNLOAD_URL}/{self.latest_tag}/{self.PACKAGE_NAME_URIS[package_name]}"
        r = requests.get(url, allow_redirects=True)
        if not r.ok:
            raise Exception(f"Error downloading Sigma release package: {r.url} {r.status_code} - {r.reason}")
        self._extract_sigma_release(r.content)

    def _extract_sigma_release(self, content: bytes):
        """Extracts the Sigma release package to the SIGMA_RULE_DIR directory.

        Args:
            content (bytes): The content of the Sigma release package.
        """

        zf = zipfile.ZipFile(io.BytesIO(content))
        zf.extractall(self.rule_dir)
        print(f"Extracted Sigma release package to {self.rule_dir}")


def get_latest_sigma_release_tag():
    """Requests https://github.com/SigmaHQ/sigma/releases/latest and returns the URL of the response
    as the latest release
    """
    url = "https://github.com/SigmaHQ/sigma/releases/latest"
    r = requests.get(url, allow_redirects=False)
    if not r.ok:
        raise Exception(f"Error getting latest Sigma release: {r.url} {r.status_code} - {r.reason}")
    latest_tag = r.next.path_url.split("/")[-1]
    if not latest_tag:
        raise Exception(f"Error getting latest Sigma release: {r.url} {r.status_code} - {r.reason}")
    return latest_tag


def setup_rule_dir(rule_dir: str) -> str:
    """Creates the SIGMA_RULE_DIR directory if it doesn't exist, and returns the path to the directory.

    Args:
        rule_dir (str): The path to a directory where Sigma rules should be installed.

    Returns:
        str: The path to the SIGMA_RULE_DIR directory.
    """
    if not rule_dir:
        rule_dir = DEFAULT_DIRS.SIGMA_RULE_DIR
    if not os.path.exists(rule_dir):
        os.makedirs(rule_dir)
    return rule_dir
